from .eventrequest import EventRequest

from .eventresponse import EventResponse

from .eventstream import \
	EventPermissionError, \
	send_event, \
	get_events, \
	get_current_event_id, \
	channel_permission_changed

from . import urls



import copy
from .storage import EventDoesNotExist
from .eventresponse import EventResponse
from .utils import make_id, publish_event, publish_kick, \
	get_storage, get_authorizer

class EventPermissionError(Exception):
	def __init__(self, message, channels=[]):
		super(Exception, self).__init__(message)
		self.channels = copy.deepcopy(channels)

def send_event(channel, event_type, data, skip_user_ids=[]):
	storage = get_storage()

	if storage:
		e = storage.append_event(channel, event_type, data)
		pub_id = str(e.id)
		pub_prev_id = str(e.id - 1)
	else:
		pub_id = None
		pub_prev_id = None

	publish_event(
		channel,
		event_type,
		data,
		pub_id,
		pub_prev_id,
		skip_user_ids=skip_user_ids)

def get_events(request, limit=100, user=None):
	resp = EventResponse()
	resp.is_next = request.is_next
	resp.is_recover = request.is_recover
	resp.user = user

	if len(request.channels) == 0:
		return resp

	limit_per_type = int(limit / len(request.channels))
	if limit_per_type < 1:
		limit_per_type = 1

	storage = get_storage()
	authorizer = get_authorizer()

	inaccessible_channels = []
	for channel in request.channels:
		if not authorizer.can_read_channel(user, channel):
			inaccessible_channels.append(channel)

	if len(inaccessible_channels) > 0:
		msg = 'Permission denied to channels: %s' % (
			', '.join(inaccessible_channels))
		raise EventPermissionError(msg, channels=inaccessible_channels)

	for channel in request.channels:
		reset = False

		last_id = request.channel_last_ids.get(channel)
		more = False

		if storage:
			if last_id is not None:
				try:
					events = storage.get_events(
						channel,
						int(last_id),
						limit=limit_per_type + 1)
					if len(events) >= limit_per_type + 1:
						events = events[:limit_per_type]
						more = True
				except EventDoesNotExist as e:
					reset = True
					events = []
					last_id = str(e.current_id)
			else:
				events = []
				last_id = str(storage.get_current_id(channel))
		else:
			events = []
			last_id = None

		resp.channel_items[channel] = events
		if last_id is not None:
			resp.channel_last_ids[channel] = last_id
		if reset:
			resp.channel_reset.add(channel)
		if more:
			resp.channel_more.add(channel)
	return resp

def get_current_event_id(channels):
	storage = get_storage()

	cur_ids = {}
	for channel in channels:
		cur_ids[channel] = str(storage.get_current_id(channel))

	return make_id(cur_ids)

def channel_permission_changed(user, channel):
	authorizer = get_authorizer()
	if not authorizer.can_read_channel(user, channel):
		user_id = user.id if user else 'anonymous'
		publish_kick(user_id, channel)



import json
import threading
import importlib
import six
from werkzeug.http import parse_options_header
from django.conf import settings
from django.http import HttpResponse
from gripcontrol import HttpStreamFormat
from django_grip import publish

try:
	from urllib import quote
except ImportError:
	from urllib.parse import quote

tlocal = threading.local()

# return dict of (channel, last-id)
def parse_grip_last(s):
	parsed = parse_options_header(s, multiple=True)

	out = {}
	for n in range(0, len(parsed), 2):
		channel = parsed[n]
		params = parsed[n + 1]
		last_id = params.get('last-id')
		if last_id is None:
			raise ValueError('channel "%s" has no last-id param' % channel)
		out[channel] = last_id
	return out

# return dict of (channel, last-id)
def parse_last_event_id(s):
	out = {}
	parts = s.split(',')
	for part in parts:
		channel, last_id = part.split(':')
		out[channel] = last_id
	return out

def make_id(ids):
	id_parts = []
	for channel, id in six.iteritems(ids):
		enc_channel = quote(channel)
		id_parts.append('%s:%s' % (enc_channel, id))
	return ','.join(id_parts)

def sse_encode_event(event_type, data, event_id=None):
	out = 'event: %s\n' % event_type
	if event_id:
		out += 'id: %s\n' % event_id
	out += 'data: %s\n\n' % json.dumps(data)
	return out

def sse_error_response(condition, text, extra={}):
	data = {'condition': condition, 'text': text}
	for k, v in six.iteritems(extra):
		data[k] = v
	body = sse_encode_event('stream-error', data, event_id='error')
	return HttpResponse(body, content_type='text/event-stream')

def publish_event(channel, event_type, data, pub_id, pub_prev_id,
		skip_user_ids=[]):
	if pub_id:
		event_id = '%I'
	else:
		event_id = None
	content = sse_encode_event(event_type, data, event_id=event_id)
	meta = {}
	if skip_user_ids:
		meta['skip_users'] = ','.join(skip_user_ids)
	publish(
		'events-%s' % channel,
		HttpStreamFormat(content),
		id=pub_id,
		prev_id=pub_prev_id,
		meta=meta)

def publish_kick(user_id, channel):
	msg = 'Permission denied to channels: %s' % channel
	data = {'condition': 'forbidden', 'text': msg, 'channels': [channel]}
	content = sse_encode_event('stream-error', data, event_id='error')
	meta = {'require_sub': 'events-%s' % channel}
	publish(
		'user-%s' % user_id,
		HttpStreamFormat(content),
		id='kick-1',
		meta=meta)
	publish(
		'user-%s' % user_id,
		HttpStreamFormat(close=True),
		id='kick-2',
		prev_id='kick-1',
		meta=meta)

def load_class(name):
	at = name.rfind('.')
	if at == -1:
		raise ValueError('class name contains no \'.\'')
	module_name = name[0:at]
	class_name = name[at + 1:]
	return getattr(importlib.import_module(module_name), class_name)()

# load and keep in thread local storage
def get_class(name):
	if not hasattr(tlocal, 'loaded'):
		tlocal.loaded = {}
	c = tlocal.loaded.get(name)
	if c is None:
		c = load_class(name)
		tlocal.loaded[name] = c
	return c

def get_class_from_setting(setting_name, default=None):
	if hasattr(settings, setting_name):
		return get_class(getattr(settings, setting_name))
	elif default:
		return get_class(default)
	else:
		return None

def get_storage():
	return get_class_from_setting('EVENTSTREAM_STORAGE_CLASS')

def get_authorizer():
	return get_class_from_setting(
		'EVENTSTREAM_AUTHORIZER_CLASS',
		'django_eventstream.authorizer.DefaultAuthorizer')



import copy
import six
from django.http import HttpResponse
from .utils import sse_encode_event, make_id

try:
	from urllib import quote
except ImportError:
	from urllib.parse import quote

class EventResponse(object):
	def __init__(self):
		self.channel_items = {}
		self.channel_last_ids = {}
		self.channel_reset = set()
		self.channel_more = set()
		self.is_next = False
		self.is_recover = False
		self.user = None

	# FIXME: use django-grip instead of manually building headers
	def to_http_response(self, http_request):
		last_ids = copy.deepcopy(self.channel_last_ids)
		event_id = make_id(last_ids)

		body = ''

		if not self.is_next:
			body += ':' + (' ' * 2048) + '\n'
			body += 'event: stream-open\ndata:\n\n'

		if len(self.channel_reset) > 0:
			body += sse_encode_event(
				'stream-reset',
				{'channels': list(self.channel_reset)},
				event_id=event_id)

		for channel, items in six.iteritems(self.channel_items):
			for item in items:
				last_ids[channel] = item.id
				event_id = make_id(last_ids)
				body += sse_encode_event(
					item.type,
					item.data,
					event_id=event_id)

		resp = HttpResponse(body, content_type='text/event-stream')

		more = (len(self.channel_more) > 0)

		params = http_request.GET.copy()
		params['link'] = 'next'
		if more:
			params['lastEventId'] = event_id
			if 'recover' in params:
				del params['recover']
		else:
			params['recover'] = 'true'
			if 'lastEventId' in params:
				del params['lastEventId']
		uri = http_request.path + '?' + params.urlencode()
		resp['Grip-Link'] = '<%s>; rel=next' % uri

		if not more:
			resp['Grip-Hold'] = 'stream'

		user_id = self.user.id if self.user else 'anonymous'

		channel_header = ''
		for channel in six.iterkeys(self.channel_items):
			if len(channel_header) > 0:
				channel_header += ', '
			enc_channel = quote(channel)
			last_id = last_ids.get(channel)
			channel_header += 'events-%s' % enc_channel
			if last_id:
				channel_header += '; prev-id=%s; filter=build-id' % last_id
			channel_header += '; filter=skip-users'
		channel_header += ', user-%s; filter=require-sub' % user_id
		resp['Grip-Channel'] = channel_header

		set_meta_header = ''

		if len(last_ids) > 0:
			id_parts = []
			for channel in six.iterkeys(last_ids):
				enc_channel = quote(channel)
				id_parts.append('%s:%%(events-%s)s' % (enc_channel, enc_channel))
			id_format = ','.join(id_parts)
			set_meta_header += 'id_format="%s"' % id_format

		if len(set_meta_header) > 0:
			set_meta_header += ', '
		set_meta_header += 'user="%s"' % user_id
		resp['Grip-Set-Meta'] = set_meta_header

		keep_alive_header = 'event: keep-alive\\ndata:\\n\\n'
		keep_alive_header += '; format=cstring; timeout=20'
		resp['Grip-Keep-Alive'] = keep_alive_header

		return resp



import sys
import json
import datetime
from django.utils import timezone
from .event import Event

is_python3 = sys.version_info >= (3,)

# minutes before purging an event from the database
EVENT_TIMEOUT = 60 * 24

# attempt to trim this many events per pass
EVENT_TRIM_BATCH = 50

class EventDoesNotExist(Exception):
	def __init__(self, message, current_id):
		super(Exception, self).__init__(message)
		self.current_id = current_id

class DjangoModelStorage(object):
	def append_event(self, channel, event_type, data):
		from . import models

		db_event = models.Event(
			channel=channel,
			type=event_type,
			data=json.dumps(data))
		db_event.save()

		self.trim_event_log()

		e = Event(
			db_event.channel,
			db_event.type,
			data,
			id=db_event.eid)

		return e

	def get_events(self, channel, last_id, limit=100):
		from . import models

		if is_python3:
			assert(isinstance(last_id, int))
		else:
			assert(isinstance(last_id, (int, long)))

		try:
			ec = models.EventCounter.objects.get(name=channel)
			cur_id = ec.value
		except models.EventCounter.DoesNotExist:
			cur_id = 0

		if last_id == cur_id:
			return []

		try:
			models.Event.objects.get(
				channel=channel,
				eid=last_id)
		except models.Event.DoesNotExist:
			raise EventDoesNotExist(
				'No such event %d' % last_id,
				cur_id)

		# increase limit by 1 since we ignore the first result
		db_events = models.Event.objects.filter(
			channel=channel,
			eid__gte=last_id
		).order_by('eid')[:limit + 1]

		if len(db_events) == 0 or db_events[0].eid != last_id:
			raise EventDoesNotExist(
				'No such event %d' % last_id,
				cur_id)

		db_events = db_events[1:]

		out = []
		for db_event in db_events:
			e = Event(
				db_event.channel,
				db_event.type,
				json.loads(db_event.data),
				id=db_event.eid)
			out.append(e)

		return out

	def get_current_id(self, channel):
		from . import models

		try:
			ec = models.EventCounter.objects.get(name=channel)
			return ec.value
		except models.EventCounter.DoesNotExist:
			return 0

	def trim_event_log(self):
		from . import models

		now = timezone.now()
		cutoff = now - datetime.timedelta(minutes=EVENT_TIMEOUT)
		while True:
			events = models.Event.objects.filter(
				created__lt=cutoff
			)[:EVENT_TRIM_BATCH]
			if len(events) < 1:
				break
			for e in events:
				try:
					e.delete()
				except models.Event.DoesNotExist:
					# someone else deleted. that's fine
					pass



import six
from .utils import parse_grip_last, parse_last_event_id

try:
	from urllib import unquote
except ImportError:
	from urllib.parse import unquote

class EventRequest(object):
	class Error(ValueError):
		pass

	class GripError(Error):
		pass

	class ResumeNotAllowedError(Error):
		pass

	def __init__(self, http_request=None, channel_limit=10):
		self.channels = set()
		self.channel_last_ids = {}
		self.is_recover = False

		if http_request:
			self.apply_http_request(http_request,
				channel_limit=channel_limit)

	def apply_http_request(self, http_request, channel_limit):
		channels = set(http_request.GET.getlist('channel'))
		is_next = False
		is_recover = False

		if len(channels) < 1:
			raise EventRequest.Error('No channels specified')

		if len(channels) > channel_limit:
			raise EventRequest.Error('Channel limit exceeded')

		if http_request.GET.get('link') == 'next':
			is_next = True

		grip_last = None
		if http_request.GET.get('recover') == 'true':
			grip_last = http_request.META.get('HTTP_GRIP_LAST')

			if grip_last:
				try:
					grip_last = parse_grip_last(grip_last)
				except:
					raise EventRequest.GripError(
						'Failed to parse Grip-Last header')
			else:
				grip_last = {}

			channel_last_ids = {}
			is_recover = True
			for grip_channel, last_id in six.iteritems(grip_last):
				if not grip_channel.startswith('events-'):
					continue
				channel = unquote(grip_channel[7:])
				if channel in channels:
					channel_last_ids[channel] = last_id
		else:
			last_event_id = http_request.META.get('HTTP_LAST_EVENT_ID')
			if not last_event_id:
				# take the first non-empty param, from the end
				for val in reversed(http_request.GET.getlist('lastEventId')):
					if val:
						last_event_id = val
						break

			if last_event_id:
				if last_event_id == 'error':
					raise EventRequest.ResumeNotAllowedError(
						'Can\'t resume session after stream-error')

				try:
					parsed = parse_last_event_id(last_event_id)

					channel_last_ids = {}
					for channel, last_id in six.iteritems(parsed):
						channel = unquote(channel)
						if channel in channels:
							channel_last_ids[channel] = last_id
				except:
					raise EventRequest.Error(
						'Failed to parse Last-Event-ID or lastEventId')
			else:
				channel_last_ids = {}

		self.channels = channels
		self.channel_last_ids = channel_last_ids
		self.is_next = is_next
		self.is_recover = is_recover



import os
import urlparse
import requests
from django.core.management.commands import runserver
from django_grip import get_pubcontrol

class Command(runserver.Command):
	help = 'Set ngrok tunnel as origin for GRIP service, then invoke runserver'

	def setup(self):
		host = None
		port = None
		ssl_host = None
		ssl_port = None

		resp = requests.get('http://localhost:4040/api/tunnels')
		tunnels = resp.json()['tunnels']
		for tunnel in tunnels:
			if tunnel['proto'] in ('http', 'https'):
				parsed = urlparse.urlparse(tunnel['public_url'])
				if tunnel['proto'] == 'http':
					host = parsed.hostname
					port = parsed.port if parsed.port is not None else 80
				elif tunnel['proto'] == 'https':
					ssl_host = parsed.hostname
					ssl_port = parsed.port if parsed.port is not None else 443

		if host is None and ssl_host is None:
			self.stderr.write('Error: no ngrok tunnels found')
			return

		pub = get_pubcontrol()
		if len(pub.clients) == 0:
			self.stderr.write('Error: no GRIP proxy configured')
			return

		pub.set_origin(
			host=host,
			port=port,
			ssl_host=ssl_host,
			ssl_port=ssl_port,
			rewrite_host=True)

		self.stdout.write(
			'Setting ngrok tunnel %s as GRIP origin' % (host or ssl_host))

	def run(self, **options):
		# be sure to execute setup() only once, even if autoreload is used
		use_reloader = options['use_reloader']
		if not use_reloader or os.environ.get('RUN_MAIN') != 'true':
			self.setup()

		super(Command, self).run(**options)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.http import HttpResponseBadRequest

def events(request):
	from .eventrequest import EventRequest
	from .eventstream import EventPermissionError, get_events
	from .utils import sse_error_response

	user = None
	if request.user.is_authenticated:
		user = request.user

	try:
		event_request = EventRequest(request)
		event_response = get_events(event_request, user=user)
		response = event_response.to_http_response(request)
	except EventRequest.ResumeNotAllowedError as e:
		response = HttpResponseBadRequest(
			'Invalid request: %s.\n' % str(e))
	except EventRequest.GripError as e:
		if request.grip_proxied:
			response = sse_error_response(
				'internal-error',
				'Invalid internal request.')
		else:
			response = sse_error_response(
				'bad-request',
				'Invalid request: %s.' % str(e))
	except EventRequest.Error as e:
		response = sse_error_response(
			'bad-request',
			'Invalid request: %s.' % str(e))
	except EventPermissionError as e:
		response = sse_error_response(
			'forbidden',
			str(e),
			{'channels': e.channels})

	response['Cache-Control'] = 'no-cache'

	if hasattr(settings, 'EVENTSTREAM_ALLOW_ORIGIN'):
		cors_origin = settings.EVENTSTREAM_ALLOW_ORIGIN
	else:
		cors_origin = request.META.get('HTTP_HOST')

	if cors_origin:
		response['Access-Control-Allow-Origin'] = cors_origin

	return response



import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
	readme = f.read()

setup(
name='django-eventstream',
version='1.1.1',
description='Django EventStream library',
long_description=readme,
author='Justin Karneges',
author_email='justin@fanout.io',
url='https://github.com/fanout/django-eventstream',
license='MIT',
zip_safe=False,
packages=['django_eventstream', 'django_eventstream.migrations', 'django_eventstream.management', 'django_eventstream.management.commands'],
package_data={'django_eventstream': ['static/django_eventstream/*']},
install_requires=['django_grip>=1.7,<3', 'Werkzeug>=0.12,<1', 'six>=1.10,<2'],
classifiers=[
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License'
]
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import IntegrityError, models, transaction

class EventCounter(models.Model):
	name = models.CharField(max_length=255, unique=True)
	value = models.BigIntegerField(default=0)
	updated = models.DateTimeField(db_index=True, auto_now=True)

	@classmethod
	def get_or_create(cls, name):
		try:
			en = cls.objects.get(name=name)
		except cls.DoesNotExist:
			try:
				en = cls(name=name)
				en.save()
			except IntegrityError:
				en = cls.objects.get(name=name)
		return en

class Event(models.Model):
	channel = models.CharField(max_length=255, db_index=True)
	type = models.CharField(max_length=255, db_index=True)
	data = models.TextField()
	eid = models.BigIntegerField(default=0, db_index=True)
	created = models.DateTimeField(db_index=True, auto_now_add=True)

	class Meta:
		unique_together = ('channel', 'eid')

	def save(self, *args, **kwargs):
		if not self.eid:
			counter = EventCounter.get_or_create(self.channel)

			with transaction.atomic():
				counter = EventCounter.objects.select_for_update(
					).get(id=counter.id)
				self.eid = counter.value + 1

				try:
					super(Event, self).save(*args, **kwargs)
				except Exception:
					self.eid = 0
					raise

				counter.value = self.eid
				counter.save()
		else:
			super(Event, self).save(*args, **kwargs)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-07-06 20:37
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('channel', models.CharField(db_index=True, max_length=255)),
                ('type', models.CharField(db_index=True, max_length=255)),
                ('data', models.TextField()),
                ('eid', models.BigIntegerField(db_index=True, default=0)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
            ],
        ),
        migrations.CreateModel(
            name='EventCounter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
                ('value', models.BigIntegerField(default=0)),
                ('updated', models.DateTimeField(auto_now=True, db_index=True)),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='event',
            unique_together=set([('channel', 'eid')]),
        ),
    ]
